from typing import Any, Callable, Optional, Set

import os
import yaml

from .constants import SNAPSHOT_DIRNAME
from .exceptions import SnapshotDoesNotExist
from .location import TestLocation
from .types import SnapshotFiles


class SnapshotIO:
    def __init__(self, test_location: TestLocation, file_hook):
        self._test_location = test_location
        self._file_hook = file_hook

    @property
    def test_location(self):
        return self._test_location

    @property
    def dirname(self) -> str:
        test_dirname = os.path.dirname(self.test_location.filename)
        snapshot_dir = self._get_snapshot_dirname()
        if snapshot_dir is not None:
            return os.path.join(test_dirname, SNAPSHOT_DIRNAME, snapshot_dir)
        return os.path.join(test_dirname, SNAPSHOT_DIRNAME)

    def discover_snapshots(self, filepath: str) -> Set[str]:
        """
        Utility method for getting all the snapshots from a file.
        Returns an empty set if the file cannot be read.
        """
        try:
            return set(self._read_file(filepath).keys())
        except:
            return set()

    def read_snapshot(self, index: int) -> Any:
        """
        Utility method for reading the contents of a snapshot assertion.
        Will call `pre_read`, then `read` and finally `post_read`,
        returning the contents parsed from the `read` method.
        """
        try:
            self.pre_read(index=index)
            return self.read(index=index)
        finally:
            self.post_read(index=index)

    def create_or_update_snapshot(self, serialized_data: Any, index: int):
        """
        Utility method for reading the contents of a snapshot assertion.
        Will call `pre_write`, then `write` and finally `post_write`.
        """
        self.pre_write(serialized_data, index=index)
        self.write(serialized_data, index=index)
        self.post_write(serialized_data, index=index)

    def delete_snapshot(self, snapshot_file: str, snapshot_name: str):
        """
        Utility method for removing a snapshot from a snapshot file.
        """
        self._write_snapshot_or_remove_file(snapshot_file, snapshot_name, None)

    def pre_read(self, index: int = 0):
        pass

    def read(self, index: int = 0) -> Any:
        snapshot_file = self.get_filepath(index)
        snapshot_name = self.get_snapshot_name(index)
        snapshot = self._read_snapshot_from_file(snapshot_file, snapshot_name)
        if snapshot is None:
            raise SnapshotDoesNotExist()
        return snapshot

    def post_read(self, index: int = 0):
        self._snap_file_hook(index)

    def pre_write(self, data: Any, index: int = 0):
        self._ensure_snapshot_dir(index)

    def write(self, data: Any, index: int = 0):
        snapshot_file = self.get_filepath(index)
        snapshot_name = self.get_snapshot_name(index)
        self._write_snapshot_or_remove_file(snapshot_file, snapshot_name, data)

    def post_write(self, data: Any, index: int = 0):
        self._snap_file_hook(index)

    def get_snapshot_name(self, index: int = 0) -> str:
        index_suffix = f".{index}" if index > 0 else ""
        methodname = self._test_location.testname

        if self._test_location.classname is not None:
            return f"{self._test_location.classname}.{methodname}{index_suffix}"
        return f"{methodname}{index_suffix}"

    def get_filepath(self, index: int) -> str:
        basename = self.get_file_basename(index=index)
        return os.path.join(self.dirname, basename)

    def get_file_basename(self, index: int) -> str:
        return f"{os.path.splitext(os.path.basename(self._test_location.filename))[0]}.yaml"

    def _get_snapshot_dirname(self) -> Optional[str]:
        return None

    def _ensure_snapshot_dir(self, index: int):
        """
        Ensures the folder path for the snapshot file exists.
        """
        try:
            os.makedirs(os.path.dirname(self.get_filepath(index)))
        except FileExistsError:
            pass

    def _read_snapshot_from_file(self, snapshot_file: str, snapshot_name: str) -> Any:
        """
        Read the snapshot file and get only the snapshot data for assertion
        """
        snapshots = self._read_file(snapshot_file)
        return snapshots.get(snapshot_name, {}).get("data", None)

    def _read_file(self, filepath: str) -> Any:
        """
        Read the snapshot data from the snapshot file into a python instance.
        """
        try:
            with open(filepath, "r") as f:
                return yaml.safe_load(f) or {}
        except FileNotFoundError:
            pass
        return {}

    def _write_snapshot_or_remove_file(
        self, snapshot_file: str, snapshot_name: str, data: Any
    ):
        """
        Adds the snapshot data to the snapshots read from the file
        or removes the snapshot entry if data is `None`.
        If the snapshot file will be empty remove the entire file.
        """
        snapshots = self._read_file(snapshot_file)
        if data is None and snapshot_name in snapshots:
            del snapshots[snapshot_name]
        else:
            snapshots[snapshot_name] = snapshots.get(snapshot_name, {})
            snapshots[snapshot_name]["data"] = data

        if snapshots:
            self._write_file(snapshot_file, snapshots)
        else:
            os.remove(snapshot_file)

    def _write_file(self, filepath: str, data: Any):
        """
        Writes the snapshot data into the snapshot file that be read later.
        """
        with open(filepath, "w") as f:
            yaml.safe_dump(data, f)

    def _snap_file_hook(self, index: int):
        """
        Notify the assertion of an access to a snapshot in a file
        """
        snapshot_file = self.get_filepath(index)
        snapshot_name = self.get_snapshot_name(index)
        self._file_hook(snapshot_file, snapshot_name)



def test_fixture(testdir):
    pyfile_content = """
def test_sth(snapshot):
    assert snapshot is not None
"""
    testdir.makepyfile(pyfile_content)

    result = testdir.runpytest("-v")
    result.stdout.fnmatch_lines(["*::test_sth PASSED*"])

    assert result.ret == 0



import os
from collections import defaultdict
from functools import lru_cache
from gettext import ngettext, gettext
from typing import Dict, List, Set, Tuple

from .assertion import SnapshotAssertion
from .constants import SNAPSHOT_DIRNAME
from .terminal import yellow, bold
from .types import SnapshotFiles


class SnapshotSession:
    def __init__(self, *, update_snapshots: bool, base_dir: str):
        self.update_snapshots = update_snapshots
        self.base_dir = base_dir
        self.discovered_snapshots: SnapshotFiles = dict()
        self.visited_snapshots: SnapshotFiles = dict()
        self.report: List[str] = []
        self._assertions: Dict[str, Dict[str, SnapshotAssertion]] = dict()

    @property
    def unused_snapshots(self) -> SnapshotFiles:
        return self._diff_snapshot_files(
            self.discovered_snapshots, self.visited_snapshots
        )

    @property
    def written_snapshots(self) -> SnapshotFiles:
        return self._diff_snapshot_files(
            self.visited_snapshots, self.discovered_snapshots
        )

    @property
    def num_unused_snapshots(self):
        return self._count_snapshots(self.unused_snapshots)

    @property
    def num_written_snapshots(self):
        return self._count_snapshots(self.written_snapshots)

    def start(self):
        self.report = []
        self.visited_snapshots = dict()
        self.discovered_snapshots = dict()

    def finish(self):
        n_unused = self.num_unused_snapshots
        n_written = self.num_written_snapshots

        self.add_report_line()

        summary_lines = []
        if self.update_snapshots and n_written:
            summary_lines += [
                ngettext(
                    "{} snapshot generated.", "{} snapshots generated.", n_written,
                ).format(bold(n_written))
            ]
        summary_lines += [
            ngettext("{} snapshot unused.", "{} snapshots unused.", n_unused).format(
                bold(n_unused)
            )
        ]
        summary_line = " ".join(summary_lines)
        self.add_report_line(
            yellow(summary_line)
            if n_unused or (self.update_snapshots and n_written)
            else summary_line
        )

        for filepath, snapshots in self.unused_snapshots.items():
            count = self._count_snapshots({filepath: snapshots})
            if not count:
                continue
            path_to_file = os.path.relpath(filepath, self.base_dir)
            self.add_report_line(
                ngettext(
                    f"{{}} at {path_to_file}",
                    f"{{}} in {path_to_file} → {', '.join(snapshots)}",
                    count,
                ).format(bold(count))
            )

        if n_unused:
            self.add_report_line()
            if self.update_snapshots:
                self.remove_unused_snapshots()
                self.add_report_line(
                    ngettext(
                        "This snapshot has been deleted.",
                        "These snapshots have been deleted.",
                        n_unused,
                    )
                )
            else:
                self.add_report_line(
                    gettext(
                        "Re-run pytest with --update-snapshots to delete the snapshots."
                    )
                )

    def add_report_line(self, line: str = ""):
        self.report += [line]

    def register_request(self, assertion: SnapshotAssertion):
        discovered = {
            filepath: assertion.io.discover_snapshots(filepath)
            for filepath in self._walk_dir(assertion.io.dirname)
        }
        self.add_discovered_snapshots(discovered)

    def register_assertion(self, assertion: SnapshotAssertion):
        filepath = assertion.io.get_filepath(assertion.num_executions)
        snapshot = assertion.io.get_snapshot_name(assertion.num_executions)
        self.add_visited_snapshots({filepath: {snapshot}})

        if filepath not in self._assertions:
            self._assertions[filepath] = dict()
        self._assertions[filepath][snapshot] = assertion

    def add_discovered_snapshots(self, snapshots: SnapshotFiles):
        self._merge_snapshot_files_into(self.discovered_snapshots, snapshots)

    def add_visited_snapshots(self, snapshots: SnapshotFiles):
        self._merge_snapshot_files_into(self.visited_snapshots, snapshots)

    def remove_unused_snapshots(self):
        for snapshot_file, unused_snapshots in self.unused_snapshots.items():
            all_discovered_unused = (
                unused_snapshots == self.discovered_snapshots[snapshot_file]
            )
            no_snapshot_written = not self.written_snapshots.get(snapshot_file)
            if all_discovered_unused and no_snapshot_written:
                os.remove(snapshot_file)
                continue
            snapshot_assertion, *_ = self._assertions[snapshot_file].values()
            for snapshot_name in unused_snapshots:
                snapshot_assertion.io.delete_snapshot(snapshot_file, snapshot_name)

    def _merge_snapshot_files_into(
        self, snapshot_files: SnapshotFiles, *snapshot_files_to_merge: SnapshotFiles,
    ):
        """
        Add snapshots from other files into the first one
        """
        for snapshot_file in snapshot_files_to_merge:
            for filepath, snapshots in snapshot_file.items():
                if self._in_snapshot_dir(filepath):
                    if filepath not in snapshot_files:
                        snapshot_files[filepath] = set()
                    snapshot_files[filepath].update(snapshots)

    def _diff_snapshot_files(
        self, snapshot_files1: SnapshotFiles, snapshot_files2: SnapshotFiles,
    ) -> SnapshotFiles:
        return {
            filename: snapshots1 - snapshot_files2.get(filename, set())
            for filename, snapshots1 in snapshot_files1.items()
        }

    def _count_snapshots(self, snapshot_files: SnapshotFiles) -> int:
        return sum(len(snaps) for snaps in snapshot_files.values())

    def _in_snapshot_dir(self, path: str) -> bool:
        parts = path.split(os.path.sep)
        return SNAPSHOT_DIRNAME in parts

    @lru_cache(maxsize=32)
    def _walk_dir(self, root: str):
        for (dirpath, _, filenames) in os.walk(root):
            if not self._in_snapshot_dir(dirpath):
                continue
            for filename in filenames:
                if not filename.startswith("."):
                    yield os.path.join(dirpath, filename)
